import * as THREE from '../build/three.module.js';
import { MeshSurfaceSampler } from './jsm/math/MeshSurfaceSampler.js';
import { GLTFLoader } from './jsm/loaders/GLTFLoader.js';
import Stats from './jsm/libs/stats.module.js';
import { GUI } from './jsm/libs/dat.gui.module.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        //renderer.outputEncoding = THREE.sRGBEncoding;
        domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        //scene.background = new THREE.Color(0xffffff);
        this._scene = scene;

        //this._setupInteractions();
        this._setupCamera();
        this._setupLight();
        this._setupGUI();
        this._setupModel();
        
        //this._setupControls();
        //this._setupHelper();
        
        window.onresize = this.resize.bind(this);

        this.resize();
    }

    _resample() {
        const count = this._api.count;
        const vertexCount = this._surface.geometry.getAttribute("position").count;

        console.info(`Sampling ${count} points from a surface with ${vertexCount} vertices...`);

        console.time(".build()");

        this._sampler = new MeshSurfaceSampler(this._surface)
            .setWeightAttribute(this._api.distribution === "weighted" ? "uv" : null)
            .build();

        console.timeEnd(".build()");

        console.time(".sample()");
        for(let i=0; i<count; i++) {
            this._ages[i] = Math.random();
            this._scales[i] = this._scaleCurve(this._ages[i]);

            this._resampleParticle(i);
        }
        console.timeEnd(".sample()");

        this._stemMesh.instanceMatrix.needsUpdate = true; 
        this._blossomMesh.instanceMatrix.needsUpdate = true;
    }

    _resampleParticle(i) {
        const position = new THREE.Vector3();
	    const normal = new THREE.Vector3();
        const dummy = new THREE.Object3D();

        this._sampler.sample(position, normal);
        normal.add(position);

        dummy.position.copy(position);
        dummy.scale.set(this._scales[i], this._scales[i], this._scales[i]);
        dummy.lookAt(normal);
        dummy.updateMatrix();

        this._stemMesh.setMatrixAt(i, dummy.matrix);
        this._blossomMesh.setMatrixAt(i, dummy.matrix);
    }

    _updateParticle(i) {
        this._ages[i] += 0.005;

        if(this._ages[i] >= 1) {
            this._ages[i] = 0.001;
            this._scales[i] = this._scaleCurve(this._ages[i]);
            this._resampleParticle(i);

            return;
        }

        const dummy = new THREE.Object3D();
        const scale = new THREE.Vector3();

        const prevScale = this._scales[i];
        this._scales[i] = this._scaleCurve(this._ages[i]);
        scale.set(this._scales[i] / prevScale, this._scales[i] / prevScale, this._scales[i] / prevScale);

        this._stemMesh.getMatrixAt(i, dummy.matrix);
        dummy.matrix.scale(scale);
        
        this._stemMesh.setMatrixAt(i, dummy.matrix);
        this._blossomMesh.setMatrixAt(i, dummy.matrix);
    }

    _setupInteractions() {
        //.
    }

    _easeOutCubic(t) {
        return (--t) * t * t + 1;
    }

    _scaleCurve(t) {
        return Math.abs(this._easeOutCubic((t>0.5 ? 1-t : t) * 2));
    } 

    _setupModel() {
        const count = this._api.count;
        this._ages = new Float32Array(count);
        this._scales = new Float32Array(count);
        
        const surfaceGeometry = new THREE.TorusKnotGeometry(10, 3, 100, 16).toNonIndexed();
        const surfaceMaterial = new THREE.MeshLambertMaterial({ color: this._api.surfaceColor, wireframe: false });
        const surface = new THREE.Mesh(surfaceGeometry, surfaceMaterial);

        const loader = new GLTFLoader();
        loader.load("./models/gltf/Flower/Flower.glb", (gltf) => {
            const stemMesh = gltf.scene.getObjectByName("Stem");
            const blossomMesh = gltf.scene.getObjectByName("Blossom");

            const stemGeometry = stemMesh.geometry.clone();
            const blossomGeometry = blossomMesh.geometry.clone();

            const defaultTransform = new THREE.Matrix4()
                .makeRotationX(Math.PI)
                .multiply(new THREE.Matrix4().makeScale(7,7,7));

            stemGeometry.applyMatrix4(defaultTransform);
            blossomGeometry.applyMatrix4(defaultTransform);

            const stemMaterial = stemMesh.material;
            const blossomMaterial = blossomMesh.material;
            
            const instancedStemMesh = new THREE.InstancedMesh(stemGeometry, stemMaterial, count);
            const instancedBlossomMesh = new THREE.InstancedMesh(blossomGeometry, blossomMaterial, count);

            const color = new THREE.Color();
            const blossomPalette = [ 0xf20587, 0xf2d479,0xf2c879, 0xf2b077, 0xf24405 ];

            for(let i=0; i<count; i++) {
                color.setHex(blossomPalette[Math.floor(Math.random() * blossomPalette.length)]);
                instancedBlossomMesh.setColorAt(i, color);
            }

            instancedStemMesh.instanceMatrix.setUsage(THREE.DynamicDrawUsage);
            instancedBlossomMesh.instanceMatrix.setUsage(THREE.DynamicDrawUsage);

            this._scene.add(surface);
            this._scene.add(instancedStemMesh);
            this._scene.add(instancedBlossomMesh);

            this._surfaceMaterial = surfaceMaterial;
            this._surface = surface;
            this._stemMesh = instancedStemMesh;
            this._blossomMesh = instancedBlossomMesh;

            this._resample();
        });
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const api = {
            count: 2000,
            distribution: "random",
            resample: this._resample.bind(this),
            surfaceColor: 0xFFF784,
            backgroundColor: 0xE39469
        };

        this._scene.background = new THREE.Color(api.backgroundColor);

        const gui = new GUI();
        gui.add(api, "count", 0, api.count).onChange(() => {
            this._stemMesh.count = api.count;
            this._blossomMesh.count = api.count;
        });

        gui.addColor(api, "backgroundColor").onChange(() => {
            this._scene.background.setHex(api.backgroundColor);
        });

        gui.addColor(api, "surfaceColor").onChange(() => {
            this._surfaceMaterial.color.setHex(api.surfaceColor);
        });

        gui.add(api, "distribution").options(["random", "weighted"]).onChange(this._resample.bind(this));
        gui.add(api, "resample");

        this._api = api;

        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);
        this._stats = stats;
    }

    _setupCamera() {    
        const camera = new THREE.PerspectiveCamera(60, window.innerWidth / window.innerHeight, 1, 100);
        camera.position.set(25, 25, 25);
        camera.lookAt(0, 0, 0);

        this._camera = camera;
    }

    _setupControls() {
        //.
    }

    _setupLight() {
        const pointLight = new THREE.PointLight(0xaa8899, 0.75);
        pointLight.position.set(50, -25, 75);
        this._scene.add(pointLight);

        this._scene.add(new THREE.HemisphereLight());
    }

    update() {
        if(this._stemMesh) {
            const time = Date.now() * 0.001;

            this._scene.rotation.x = Math.sin(time / 4);
            this._scene.rotation.y = Math.sin(time / 2);

            for(let i=0; i<this._api.count; i++) {
                this._updateParticle(i);
            }

            this._stemMesh.instanceMatrix.needsUpdate = true;
            this._blossomMesh.instanceMatrix.needsUpdate = true;
        }

        this._stats.update();
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();

        this._renderer.setPixelRatio( window.devicePixelRatio );
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    (new App()).render();
}